# Changelog

{{#each releases}}
  {{#if href}}
    ## [{{title}}]({{href}})
  {{else}}
    ## {{title}}
  {{/if}}

  {{#if tag}}
    > {{niceDate}}
  {{/if}}

  {{#if summary}}
    {{summary}}
  {{/if}}

  {{#each merges}}
    - {{#if commit.breaking}}**Breaking change:** {{/if}}{{message}}{{#if href}} [#{{id}}]({{href}}){{/if}}
  {{/each}}
  {{#each fixes}}
    - {{#if commit.breaking}}**Breaking change:** {{/if}}{{commit.subject}}{{#each fixes}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}}){{/each}}
  {{/each}}
  {{#each commits}}
    - {{#if breaking}}**Breaking change:** {{/if}}{{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
  {{/each}}

{{/each}}
